import collections
import logging
import os
import sys

from datarobot_batch_scoring.consts import (WriterQueueMsg, Batch,
                                            ProgressQueueMsg)
from datarobot_batch_scoring.reader import (fast_to_csv_chunk,
                                            slow_to_csv_chunk)
from datarobot_batch_scoring.utils import compress, get_rusage, Worker


logger = logging.getLogger(__name__)


FakeResponse = collections.namedtuple('FakeResponse', 'status_code, text')


class BaseNetworkWorker(Worker):
    """A network worker

    Work for the worker is read off of the network_queue; failures are put into
    the network_deque from which they are also read and sometimes retried.

    Successfully returned responses are put into the writer_queue for
    processing by another worker.

    Occasional progress updates are occasionally put into the progress_queue
    for processing by another worker

    A properly functioning Worker implementation must read data from the
    network_queue until it finds the SENTINEL message, which is the indicator
    that no more work will be put onto the network queue. At this point the
    worker should finish any remaining work and put a NETWORK_PROGRESS message
    into the progress_queue with the following details:

    ret : boolean
        The return status (success flag) of the process
    processed : int
        The number of requests processed
    retried : int
        The number of requests retried (read from the network_deque)
    consumed : int
        The number of consumed requests from the network_queue
    rusage : dict
        The resource usage. See the function `get_rusage` in `utils`

    In addition, it is helpful for debugging purposes if the worker is diligent
    about minding its state transitions. As an instance of `Worker`, any time
    the `self.state` attribute is updated, a message is sent to the interface
    logging this fact. Valid values for the state are defined in this class in
    the `state_names` class attribute.
    """

    state_names = {
        b"-": "Initial",
        b"I": "Idle",
        b"e": "PreIdle",
        b"E": "PrePreIdle",
        b"R": "Doing Requests",
        b"F": "Pool is Full",
        b"W": "Waiting for Finish",
        b"D": "Done"
    }

    def __init__(self, concurrency, timeout, ui,
                 network_queue,
                 network_deque,
                 writer_queue,
                 progress_queue,
                 abort_flag,
                 network_status,
                 endpoint,
                 headers,
                 user,
                 api_token,
                 pred_name,
                 fast_mode,
                 max_batch_size,
                 compression,
                 verify_ssl):

        Worker.__init__(self, network_status)

        self.concurrency = concurrency
        self.timeout = timeout
        self.ui = ui or logger
        self.network_queue = network_queue
        self.network_deque = network_deque
        self.writer_queue = writer_queue
        self.progress_queue = progress_queue
        self.abort_flag = abort_flag
        self.endpoint = endpoint
        self.headers = headers
        self.user = user
        self.api_token = api_token
        self.pred_name = pred_name
        self.fast_mode = fast_mode
        self.max_batch_size = max_batch_size
        self.compression = compression
        self.verify_ssl = verify_ssl

        self._timeout = timeout
        self.futures = []
        self.concurrency = concurrency

        self._executor = None
        self.session = None
        self.proc = None

        self.n_consumed = 0
        self.n_retried = 0
        self.n_requests = 0

    def send_warning_to_ctx(self, batch, message):
        self.ui.info('CTX WARNING batch_id {} , '
                     'message {}'.format(batch.id, message))
        self.writer_queue.put((WriterQueueMsg.CTX_WARNING, {
            "batch": batch,
            "error": message
        }))

    def send_error_to_ctx(self, batch, message):
        self.ui.info('CTX ERROR batch_id {} , '
                     'message {}'.format(batch.id, message))

        self.writer_queue.put((WriterQueueMsg.CTX_ERROR, {
            "batch": batch,
            "error": message
        }))

    def split_batch(self, batch):
        if self.fast_mode:
            chunk_formatter = fast_to_csv_chunk
        else:
            chunk_formatter = slow_to_csv_chunk

        todo = [batch]
        while todo:
            batch = todo.pop(0)
            data = chunk_formatter(batch.data, batch.fieldnames)
            starting_size = sys.getsizeof(data)
            if starting_size < self.max_batch_size:
                if self.compression:
                    data = compress(data)
                    self.ui.debug(
                        'batch {}-{} transmitting {} byte - space savings '
                        '{}%'.format(batch.id, batch.rows,
                                     sys.getsizeof(data),
                                     '%.2f' % float(1 -
                                                    (sys.getsizeof(data) /
                                                     starting_size))))
                else:
                    self.ui.debug('batch {}-{} transmitting {} bytes'
                                  ''.format(batch.id, batch.rows,
                                            starting_size))

                yield (batch, data)
            else:
                if batch.rows < 2:
                    msg = ('batch {} is single row but bigger '
                           'than limit, skipping. We lost {} '
                           'records'.format(batch.id,
                                            len(batch.data)))
                    self.ui.error(msg)
                    self.send_error_to_ctx(batch, msg)
                    continue

                msg = ('batch {}-{} is too long: {} bytes,'
                       ' splitting'.format(batch.id, batch.rows,
                                           len(data)))
                self.ui.debug(msg)
                self.send_warning_to_ctx(batch, msg)
                split_point = int(batch.rows/2)

                data1 = batch.data[:split_point]
                batch1 = Batch(batch.id, split_point, batch.fieldnames,
                               data1, batch.rty_cnt)
                todo.append(batch1)

                data2 = batch.data[split_point:]
                batch2 = Batch(batch.id + split_point,
                               batch.rows - split_point,
                               batch.fieldnames, data2, batch.rty_cnt)
                todo.append(batch2)
                todo.sort()

    def exit_fast(self, a, b):
        self.state = b'D'
        os._exit(1)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.proc and self.proc.is_alive():
            self.proc.terminate()

    def run(self):
        """This is a dummy implementation. It will return immediately."""
        self.progress_queue.put((ProgressQueueMsg.NETWORK_DONE, {
            "ret": True,
            "processed": self.n_requests,
            "retried": self.n_retried,
            "consumed": self.n_consumed,
            "rusage": get_rusage(),
        }))

    def go(self):
        """This is a dummy implementation. It will return immediately."""
        return self.run()



import collections
import logging
import signal

from concurrent.futures import FIRST_COMPLETED
from concurrent.futures import wait
from datarobot_batch_scoring.consts import (SENTINEL)
from six.moves import queue

from .base_network_worker import BaseNetworkWorker

logger = logging.getLogger(__name__)


FakeResponse = collections.namedtuple('FakeResponse', 'status_code, text')


class DryRunNetworkWorker(BaseNetworkWorker):
    """A worker that will drain the network_queue, but doesn't actually send
    any requests or put anything into the writer_queue
    """

    def get_batch(self):
        while True:
            if self.abort_flag.value:
                self.exit_fast(None, None)
                break
            try:
                r = self.network_deque.get_nowait()
                self.ui.debug('Got batch from dequeu: {}'.format(r.id))
                self.n_retried += 1
                yield r
            except queue.Empty:
                try:
                    r = self.network_queue.get(timeout=1)
                    if r.id == SENTINEL.id:
                        break
                    self.n_consumed += 1
                    yield r
                except queue.Empty:
                    if self.state in (b"-", b"R"):
                        self.state = b'E'
                    elif self.state == b"E":
                        self.state = b'e'
                    elif self.state == b"e":
                        self.state = b'I'
                        break

                except OSError:
                    self.ui.error('OS Error')
                    break

    def perform_requests(self):
        signal.signal(signal.SIGINT, self.exit_fast)
        signal.signal(signal.SIGTERM, self.exit_fast)

        self.state = b'E'
        for q_batch in self.get_batch():
            for (_, _) in self.split_batch(q_batch):
                if self.state != b"R":
                    self.state = b'R'
                yield
                continue

        #  wait for all batches to finish before returning
        self.state = b'W'
        while self.futures:
            f_len = len(self.futures)
            self.futures = [i for i in self.futures if not i.done()]
            if f_len != len(self.futures):
                self.ui.debug('Waiting for final requests to finish. '
                              'remaining requests: {}'
                              ''.format(len(self.futures)))
            wait(self.futures, return_when=FIRST_COMPLETED)
        self.state = b'D'
        yield True

    def run(self):
        i = 0
        for _ in self.perform_requests():
            i += 1
        return i

    def go(self):
        return self.run()



import codecs
import csv
import getpass
import gzip
import io
import logging
import os
import sys
from contextlib import contextmanager
from collections import namedtuple
from functools import partial
from gzip import GzipFile
from os import getcwd
from os.path import expanduser, isfile, join as path_join

import requests
import six
import trafaret as t
from six.moves.configparser import ConfigParser
from six.moves import input

if six.PY2:
    from urlparse import urlparse
elif six.PY3:
    from urllib.parse import urlparse

OptKey = partial(t.Key, optional=True)


CONFIG_FILENAME = 'batch_scoring.ini'


def verify_objectid(value):
    """Verify if id_ is a proper ObjectId. """
    try:
        t.Regexp(regexp='^[A-Fa-f0-9]{24}$').check(value)
    except t.DataError:
        raise ValueError('id {} not a valid project/model id'.format(value))


config_validator = t.Dict({
    OptKey('host'): t.String,
    OptKey('project_id'): t.Regexp(regexp='^[A-Fa-f0-9]{24}$'),
    OptKey('model_id'): t.Regexp(regexp='^[A-Fa-f0-9]{24}$'),
    OptKey('deployment_id'): t.Regexp(regexp='^[A-Fa-f0-9]{24}$'),
    OptKey('import_id'): t.String,
    OptKey('n_retry'): t.Int,
    OptKey('keep_cols'): t.String,
    OptKey('n_concurrent'): t.Int,
    OptKey('dataset'): t.String,
    OptKey('n_samples'): t.Int,
    OptKey('delimiter'): t.String,
    OptKey('out'): t.String,
    OptKey('user'): t.String,
    OptKey('password'): t.String,
    OptKey('datarobot_key'): t.String,
    OptKey('timeout'): t.Int,
    OptKey('api_token'): t.String,
    OptKey('create_api_token'): t.String,
    OptKey('pred_name'): t.String,
    OptKey('skip_row_id'): t.StrBool,
    OptKey('output_delimiter'): t.String,
    OptKey('field_size_limit'): t.Int,
    OptKey('ca_bundle'): t.String,
    OptKey('no_verify_ssl'): t.StrBool,
    OptKey('max_prediction_explanations'): t.Int,
}).allow_extra('*')


logger = logging.getLogger('main')
root_logger = logging.getLogger()


class UI(object):

    def __init__(self, prompt, loglevel, stdout, file_name_suffix='main'):
        self._prompt = prompt
        self.loglevel = loglevel
        self.stdout = stdout
        self.log_files = []
        self.file_name_suffix = file_name_suffix
        self._configure_logging(loglevel, stdout)

    def _configure_logging(self, level, stdout):
        """Configures logging for user and debug logging. """
        self.root_logger_filename = self.get_file_name(self.file_name_suffix)
        if isfile(self.root_logger_filename):
            os.unlink(self.root_logger_filename)
        if self.file_name_suffix is 'main':
            self.log_files.append(str(self.root_logger_filename))

        # user logger
        if self.file_name_suffix != 'main':
            fs = '%(processName)s %(asctime)-15s [%(levelname)s] %(message)s'
            hdlr = logging.FileHandler(self.root_logger_filename, 'w+')
        elif stdout:
            fs = '%(processName)s [%(levelname)s] %(message)s'
            hdlr = logging.StreamHandler(sys.stdout)
        else:
            fs = '%(processName)s [%(levelname)s] %(message)s'
            hdlr = logging.StreamHandler()
        dfs = None
        fmt = logging.Formatter(fs, dfs)
        hdlr.setFormatter(fmt)
        logger.setLevel(level)
        logger.addHandler(hdlr)

        # root logger
        if stdout is False and self.file_name_suffix == 'main':
            fs = '%(processName)s %(asctime)-15s [%(levelname)s] %(message)s'
            hdlr = logging.FileHandler(self.root_logger_filename, 'w+')
            dfs = None
            fmt = logging.Formatter(fs, dfs)
            hdlr.setFormatter(fmt)
            root_logger.setLevel(logging.DEBUG)
            root_logger.addHandler(hdlr)

    def prompt_yesno(self, msg):
        if self._prompt is not None:
            return self._prompt
        cmd = input('{} (Yes/No)> '.format(msg)).strip().lower()
        while cmd not in ('yes', 'no', 'y', 'n'):
            cmd = input('Please type (Yes/No)> ').strip().lower()
        return cmd in ('yes', 'y')

    def prompt_user(self):
        return input('user name> ').strip()

    def debug(self, msg):
        logger.debug(msg)

    def info(self, msg):
        logger.info(msg)

    def warning(self, msg):
        logger.warning(msg)

    def error(self, msg):
        if sys.exc_info()[0]:
            exc_info = True
        else:
            exc_info = False
        if self.file_name_suffix != 'main':
            logger.error(msg, exc_info=exc_info)
        elif self.stdout:
            logger.error(msg, exc_info=exc_info)
        else:
            logger.error(msg)
            root_logger.error(msg, exc_info=exc_info)

    def fatal(self, msg):
        exc_info = sys.exc_info()
        if self.file_name_suffix != 'main':
            msg = ('{}\nIf you need assistance please send the log file/s:\n'
                   '{}to support@datarobot.com.').format(
                           msg, self.get_all_logfiles())
            logger.error(msg, exc_info=exc_info)
        elif self.stdout:
            msg = ('{}\nIf you need assistance please send the output of this '
                   'script to support@datarobot.com.').format(
                   msg)
            logger.error(msg, exc_info=exc_info)
        else:
            msg = ('{}\nIf you need assistance please send the log file/s:\n'
                   '{}to support@datarobot.com.').format(
                           msg, self.get_all_logfiles())
            logger.error(msg)
            root_logger.error(msg, exc_info=exc_info)
        self.close()
        sys.exit(1)

    def close(self):
        for l in [logger, root_logger]:
            handlers = l.handlers[:]
            for h in handlers:
                if hasattr(h, 'close'):
                    h.close()
                l.removeHandler(h)
            if hasattr(l, 'shutdown'):
                l.shutdown()

    def get_file_name(self, suffix):
        return os.path.join(os.getcwd(), 'datarobot_batch_scoring_{}.log'
                                         ''.format(suffix))

    def get_all_logfiles(self):
        file_names = ''
        for logfile in self.log_files:
            file_names += '\t{}\n'.format(logfile)
        return file_names

    def set_next_UI_name(self, suffix):
        """
        On the Windows platform we want to init a new UI inside each subproc
        This allows us to set the name new log file name after pickling
        On *NIX we can just write to a single file
        """
        if self.file_name_suffix != 'main':
            #  For now let's only init new UI's from the main process
            self.error('set_next_UI_name() called by "{}" UI instance. This '
                       'should only be called by "main".'
                       ''.format(self.file_name_suffix))
        if os.name is 'nt':
            self._next_suffix = suffix
            self.log_files.append(self.get_file_name(suffix))
        else:
            self._next_suffix = 'main'
            self.log_files.append(self.get_file_name(self._next_suffix))
            self.log_files = list(set(self.log_files))  # dedupe list

    def __getstate__(self):
        """
        On windows we need to pickle the UI instances or create new
        instances inside the subprocesses since there's no fork.
        """
        d = self.__dict__.copy()
        #  replace the old file suffix with a separate log file
        d['file_name_suffix'] = self._next_suffix
        #  remove args not needed for __new__
        for k in [i for i in d.keys()]:
            if k not in ['_prompt', 'loglevel', 'stdout', 'file_name_suffix']:
                del d[k]
        return d

    def __setstate__(self, d):
        """
        On windows we need to pickle the UI instances or create new
        instances inside the subprocesses since there's no fork.
        This method is called when unpickling a UI instance.
        It actually creates a new UI that logs to a separate file.
        """
        if os.name is not 'nt':
            raise SystemError('__getstate__() should not be called in '
                              'non-windows environments.')
        self.__dict__.update(d)
        self._configure_logging(self.loglevel, self.stdout)

    def getpass(self):
        if self._prompt is not None:
            raise RuntimeError("Non-interactive session")
        return getpass.getpass('password> ')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def get_config_file():
    """
    Lookup for config file at user home directory or working directory.
    Returns
    -------
    str or None
        Path to config file or None if it not exists.
    """
    home_path = path_join(expanduser('~'), CONFIG_FILENAME)
    cwd_path = path_join(getcwd(), CONFIG_FILENAME)
    if isfile(cwd_path):
        return cwd_path
    elif isfile(home_path):
        return home_path
    return None


def parse_config_file(file_path):
    config = ConfigParser()
    config.read(file_path)
    if 'batch_scoring' not in config.sections():
        #  We are return empty dict, because there is nothing in this file
        #  that related to arguments to batch scoring.
        return {}
    parsed_dict = dict(config.items('batch_scoring'))
    return config_validator(parsed_dict)


def acquire_api_token(base_url, base_headers, user, pwd,
                      create_api_token, ui, verify_ssl):
    """Get the api token.

    Either supplied by user or requested from the API with username and pwd.
    Optionally, create a new one.
    """

    auth = (user, pwd)

    if create_api_token:
        request_meth = requests.post
    else:
        request_meth = requests.get

    r = request_meth(base_url + 'api_token',
                     auth=auth,
                     headers=base_headers,
                     verify=verify_ssl)

    if r.status_code == 401:
        raise ValueError('wrong credentials')
    elif r.status_code != 200:
        raise ValueError('api_token request returned status code {}'
                         .format(r.status_code))
    else:
        ui.info('api-token acquired')

    api_token = r.json().get('api_token')

    if api_token is None:
        raise ValueError('no api-token registered; '
                         'please run with --create_api_token flag.')

    ui.debug('api-token: {}'.format(api_token))

    return api_token


def parse_host(host, ui):
    parsed = urlparse(host)
    ui.debug('urlparse.urlparse result: {}'.format(parsed))
    if not parsed.scheme and not parsed.netloc:
        #  protocol missing
        ui.fatal('Cannot parse "--host" argument. Host address must start '
                 'with a protocol such as "http://" or "https://". '
                 'Value given: {}'.format(host))
    base_url = '{}://{}/predApi/v1.0/'.format(parsed.scheme,
                                              parsed.netloc)
    ui.debug('parse_host return value: {}'.format(base_url))
    return base_url


def compress(data):
    buf = io.BytesIO()
    with GzipFile(fileobj=buf, mode='wb', compresslevel=2) as f:
        f.write(data)
    return buf.getvalue()


def warn_if_redirected(req, ui):
    """
    test whether a request was redirect.
    Log a warning to the user if it was redirected
    """
    if req is None:
        return
    history = req.history
    if history:
        first = history[0]
        if first.is_redirect:
            starting_endpoint = first.url  # Requested url
            redirect_endpoint = first.headers.get('Location')  # redirect
            if str(starting_endpoint) != str(redirect_endpoint):
                ui.warning('The requested URL:\n\t{}\n\twas redirected '
                           'by the webserver to:\n\t{}'
                           ''.format(starting_endpoint, redirect_endpoint))


def get_endpoint(host, api_version):
    parsed = urlparse(host)
    if not parsed.scheme and not parsed.netloc:
        raise ValueError(
            'Cannot parse "--host" argument. Host address must start '
            'with a protocol such as "http://" or "https://". '
            'Value given: {}'.format(host))

    return '{}://{}/{}/'.format(parsed.scheme, parsed.netloc, api_version)


def make_validation_call(user, api_token, n_retry, endpoint, base_headers,
                         batch, ui, compression=None, verify_ssl=True):
    """Check if user is authorized for the given model and that schema is
    correct.

    This function will make a sync request to the api endpoint with a single
    row just to make sure that the schema is correct and the user
    is authorized.
    """
    r = None

    while n_retry:
        ui.debug('request authorization')
        if compression:
            data = compress(batch.data)
        else:
            data = batch.data
        try:
            if user and api_token:
                r = requests.post(endpoint, headers=base_headers,
                                  data=data,
                                  auth=(user, api_token),
                                  verify=verify_ssl)
            elif not (user and api_token):
                r = requests.post(endpoint, headers=base_headers,
                                  data=data,
                                  verify=verify_ssl)
            else:
                ui.fatal("Aborting: no auth credentials passed")

            ui.debug(u'authorization request response: {}|{}'
                     .format(r.status_code, r.text))
            if r.status_code == 200:
                # all good
                break

            warn_if_redirected(r, ui)
            if r.status_code == 400:
                # client error -- maybe schema is wrong
                try:
                    msg = r.json()['message']
                except:
                    msg = r.text
                ui.fatal(u'Failed with client error: {}.'.format(msg))
            elif r.status_code == 403:
                #  This is usually a bad API token. E.g.
                #  {"status": "API token not valid", "code": 403}
                ui.fatal(u'Failed with message:\n\t{}.'.format(r.text))
            elif r.status_code == 401:
                #  This can be caused by having the wrong datarobot_key
                ui.fatal('Failed to authenticate -- '
                         'Please check your: datarobot_key (if required), '
                         'username/password and/or api token. Contact '
                         'customer support if the problem persists '
                         'message:\n{}.'
                         ''.format(r.__dict__.get('_content')))
            elif r.status_code == 405:
                ui.fatal('Failed to request endpoint -- Please check your '
                         '"--host" argument.')
            elif r.status_code == 502:
                ui.fatal('Problem with the gateway -- Please check your '
                         '"--host" argument and contact customer support'
                         'if the problem persists.')
            elif r.status_code == 422:
                try:
                    msg = r.json()['message']
                except:
                    msg = r.text
                ui.fatal(u'Predictions are not available '
                         'because: \"{}\".'.format(msg))
        except requests.exceptions.SSLError as e:
            ui.error('SSL verification failed, reason: {}.'.format(e))
        except requests.exceptions.ConnectionError:
            ui.error('Cannot connect to {}'.format(endpoint))
        n_retry -= 1

    if n_retry == 0:
        status = r.text if r is not None else 'UNKNOWN'
        try:
            status = r.json()['message']
        except:
            pass  # fall back to r.text
        content = r.content if r is not None else 'NO CONTENT'
        warn_if_redirected(r, ui)
        ui.debug("Failed authorization response \n{!r}".format(content))
        ui.fatal('Authorization failed -- Please check project id and model '
                 'id permissions: {}.'.format(status))
    else:
        ui.debug('Authorization has succeeded.')


try:
    import resource

    def get_rusage():
        usage = resource.getrusage(resource.RUSAGE_SELF)
        return {
            "utime": usage.ru_utime,
            "stime": usage.ru_stime,
            "rss": usage.ru_maxrss,
        }

except ImportError:
    def get_rusage():
        return {}


DialectTuple = namedtuple('DialectTuple',
                          'delimiter doublequote escapechar lineterminator '
                          'quotechar quoting skipinitialspace strict')


class SerializableDialect(DialectTuple):
    @classmethod
    def from_dialect(cls, dialect):
        return cls(
            delimiter=dialect.delimiter,
            doublequote=dialect.doublequote,
            escapechar=dialect.escapechar,
            lineterminator=dialect.lineterminator,
            quotechar=dialect.quotechar,
            quoting=dialect.quoting,
            skipinitialspace=dialect.skipinitialspace,
            strict=getattr(dialect, 'strict', False))

    def to_dialect(self):
        class _dialect(csv.Dialect):
            delimiter = self.delimiter
            doublequote = self.doublequote
            escapechar = self.escapechar
            lineterminator = self.lineterminator
            quotechar = self.quotechar
            quoting = self.quoting
            skipinitialspace = self.skipinitialspace
            strict = self.strict
        return _dialect()


class Worker(object):
    state_names = {}

    def __init__(self, status_value):
        self.status_value = status_value

    @property
    def state(self):
        return self.status_value.value

    @state.setter
    def state(self, status):
        self.ui.debug('state: {} -> {}'
                      ''.format(self.state_name(), self.state_name(status)))
        self.status_value.value = status

    def state_name(self, s=None):
        return self.state_names[s or self.state]


@contextmanager
def gzip_with_encoding(data, mode, encoding=None):
    """ Decorator to support encoding for gzip in PY2
    """
    if encoding is not None:
        reader = codecs.getreader(encoding)
        with gzip.open(data, mode) as f:
            yield reader(f)
    else:
        with gzip.open(data, mode) as f:
            yield f



from . import pred_api_v10, api_v1

PRED_API_V10 = 'predApi/v1.0'
API_V1 = 'api/v1'

RESPONSE_HANDLERS = {
    PRED_API_V10: (pred_api_v10.unpack_data, pred_api_v10.format_data),
    API_V1: (api_v1.unpack_data, api_v1.format_data),
}


def get_response_handlers_from_url(url):
    if PRED_API_V10 in url:
        return RESPONSE_HANDLERS[PRED_API_V10]
    elif API_V1 in url:
        return RESPONSE_HANDLERS[API_V1]

    raise ValueError('Failed to find a handler for url: {}'.format(url))



import operator
import json

from datarobot_batch_scoring.consts import TargetType
from datarobot_batch_scoring.exceptions import UnexpectedKeptColumnCount


def format_data(result, batch, **opts):
    pred_name = opts.get('pred_name')
    keep_cols = opts.get('keep_cols')
    skip_row_id = opts.get('skip_row_id')
    fast_mode = opts.get('fast_mode')

    predictions = result['predictions']
    if result['task'] == TargetType.BINARY:
        sorted_classes = list(
            sorted(predictions[0]['class_probabilities'].keys()))
        out_fields = ['row_id'] + sorted_classes
        if pred_name is not None:
            sorted_classes = [sorted_classes[-1]]
            out_fields = ['row_id'] + [pred_name]
        pred = [[p['row_id'] + batch.id] +
                [p['class_probabilities'][c] for c in
                 sorted_classes]
                for p in
                sorted(predictions,
                       key=operator.itemgetter('row_id'))]
    elif result['task'] == TargetType.REGRESSION:
        pred = [[p['row_id'] + batch.id, p['prediction']]
                for p in
                sorted(predictions,
                       key=operator.itemgetter('row_id'))]
        out_fields = ['row_id', pred_name if pred_name else '']
    else:
        raise ValueError('task "{}" not supported'
                         ''.format(result['task']))

    if keep_cols:
        # stack columns

        feature_indices = {col: i for i, col in
                           enumerate(batch.fieldnames)}
        indices = [feature_indices[col] for col in keep_cols]

        written_fields = []

        if not skip_row_id:
            written_fields.append('row_id')

        written_fields += keep_cols + out_fields[1:]

        # first column is row_id
        comb = []
        for row, predicted in zip(batch.data, pred):
            if fast_mode:
                # row is a full line, we need to cut it into fields
                if len(row) != len(batch.fieldnames):
                    raise UnexpectedKeptColumnCount()
            keeps = [row[i] for i in indices]
            output_row = []
            if not skip_row_id:
                output_row.append(predicted[0])
            output_row += keeps + predicted[1:]
            comb.append(output_row)
    else:
        if not skip_row_id:
            comb = pred
            written_fields = out_fields
        else:
            written_fields = out_fields[1:]
            comb = [row[1:] for row in pred]

    return written_fields, comb


def unpack_data(request):
    result = json.loads(request['text'])
    exec_time = result['execution_time']
    elapsed_total_seconds = request['elapsed']
    return result, exec_time, elapsed_total_seconds



import csv
import glob
import hashlib
import multiprocessing
import operator
import os
import shelve
import signal
import sys
from functools import reduce
from time import time

import six
from six.moves import queue

from datarobot_batch_scoring.consts import SENTINEL, \
    WriterQueueMsg, ProgressQueueMsg, REPORT_INTERVAL
from datarobot_batch_scoring.utils import get_rusage
from datarobot_batch_scoring.exceptions import ShelveError, \
    UnexpectedKeptColumnCount, NoPredictionThresholdInResult

if six.PY3:
    import dbm.dumb as dumb_dbm
else:
    import dumbdbm as dumb_dbm


class RunContext(object):
    """A context for a run backed by a persistant store.

    We use a shelve to store the state of the run including
    a journal of processed batches that have been checkpointed.

    Note: we use globs for the shelve files because different
    versions of Python have different file layouts.
    """

    def __init__(self, n_samples, out_file, pid, lid, keep_cols,
                 n_retry, delimiter, dataset, pred_name, ui, file_context,
                 fast_mode, encoding, skip_row_id, output_delimiter,
                 pred_threshold_name, pred_decision_name,
                 max_prediction_explanations,
                 ):
        self.n_samples = n_samples
        self.out_file = out_file
        self.project_id = pid
        self.model_id = lid
        self.keep_cols = keep_cols
        self.n_retry = n_retry
        self.delimiter = delimiter
        self.dataset = dataset
        self.pred_name = pred_name
        self.pred_threshold_name = pred_threshold_name
        self.pred_decision_name = pred_decision_name
        self.out_stream = None
        self._ui = ui
        self.file_context = file_context
        self.fast_mode = fast_mode
        self.encoding = encoding
        self.skip_row_id = skip_row_id
        self.output_delimiter = output_delimiter
        #  dataset_dialect and writer_dialect are set by
        #  investigate_encoding_and_dialect in utils
        self.dialect = csv.get_dialect('dataset_dialect')
        self.writer_dialect = csv.get_dialect('writer_dialect')
        self.scoring_succeeded = False  # Removes shelves when True
        self.is_open = False  # Removes shelves when True
        self.max_prediction_explanations = max_prediction_explanations

    @classmethod
    def create(cls, resume, n_samples, out_file, pid, lid,
               keep_cols, n_retry,
               delimiter, dataset, pred_name, ui,
               fast_mode, encoding, skip_row_id, output_delimiter,
               pred_threshold_name, pred_decision_name,
               max_prediction_explanations,
               ):
        """Factory method for run contexts.

        Either resume or start a new one.
        """
        file_context = ContextFile(pid, lid, n_samples, keep_cols)
        if resume is None and file_context.exists():
            resume = ui.prompt_yesno('Existing run found. Resume')
        if resume:
            ctx_class = OldRunContext
        else:
            ctx_class = NewRunContext

        return ctx_class(n_samples, out_file, pid, lid, keep_cols, n_retry,
                         delimiter, dataset, pred_name, ui, file_context,
                         fast_mode, encoding, skip_row_id, output_delimiter,
                         pred_threshold_name, pred_decision_name,
                         max_prediction_explanations)

    def __enter__(self):
        assert(not self.is_open)
        self.is_open = True
        self._ui.debug('ENTER CALLED ON RUNCONTEXT')
        raw_db = dumb_dbm.open(self.file_context.file_name)
        self.db = shelve.Shelf(raw_db, writeback=True)
        if not hasattr(self, 'partitions'):
            self.partitions = []
        return self

    def __exit__(self, type, value, traceback):
        if not self.is_open:
            self._ui.debug('EXIT CALLED ON CLOSED RUNCONTEXT: successes={}'
                           ''.format(self.scoring_succeeded))
            return
        self.is_open = False
        self._ui.debug('EXIT CALLED ON RUNCONTEXT: successes={}'
                       ''.format(self.scoring_succeeded))
        self.db.close()
        if self.out_stream is not None:
            self.out_stream.close()
        if self.scoring_succeeded:
            # success - remove shelve
            self.file_context.clean()

    def checkpoint_batch(self, batch, written_fields, comb):
        """Mark a batch as being processed:
           - write it to the output stream (if necessary pull out columns).
           - put the batch_id into the journal.
        """

        # if an error happens during/after the append we
        # might end up with inconsistent state
        # TODO write partition files instead of appending
        #  store checksum of each partition and back-check
        writer = csv.writer(self.out_stream, dialect=self.writer_dialect)
        if self.db['first_write']:
            writer.writerow(written_fields)
        writer.writerows(comb)
        self.out_stream.flush()

        self.db['checkpoints'].append((batch.id, batch.rows))

        self.db['first_write'] = False
        self._ui.info('batch {}-{} checkpointed'
                      .format(batch.id, batch.rows))
        self.db.sync()

    def save_error(self, batch, error, bucket="errors"):
        msgs = self.db[bucket].setdefault((batch.id, batch.rows), [])
        msgs.append(error)
        self.db.sync()

    def save_warning(self, batch, error):
        self.save_error(batch, error, "warnings")

    def __getstate__(self):
        assert(not self.is_open)
        self.out_stream = None
        d = self.__dict__.copy()
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.open()

    def open(self):
        if self.is_open:
            self._ui.debug('OPEN CALLED ON ALREADY OPEN RUNCONTEXT')
            return
        self.is_open = True
        self._ui.debug('OPEN CALLED ON RUNCONTEXT')
        csv.register_dialect('dataset_dialect', **self.dialect)
        csv.register_dialect('writer_dialect', **self.writer_dialect)
        self.dialect = csv.get_dialect('dataset_dialect')
        self.writer_dialect = csv.get_dialect('writer_dialect')
        self.db = shelve.open(self.file_context.file_name, writeback=True)
        if six.PY2:
            self.out_stream = open(self.out_file, 'ab')
        elif six.PY3:
            self.out_stream = open(self.out_file, 'a', newline='')

    def close(self):
        if not self.is_open:
            self._ui.debug('CLOSE CALLED ON CLOSED RUNCONTEXT')
            return
        self.is_open = False
        self._ui.debug('CLOSE CALLED ON RUNCONTEXT')
        self.dialect = csv.get_dialect('dataset_dialect')
        self.writer_dialect = csv.get_dialect('writer_dialect')
        values = ['delimiter', 'doublequote', 'escapechar', 'lineterminator',
                  'quotechar', 'quoting', 'skipinitialspace', 'strict']
        self.dialect = {k: getattr(self.dialect, k) for k in values if
                        hasattr(self.dialect, k)}
        self.writer_dialect = {k: getattr(self.writer_dialect, k) for k
                               in values if hasattr(self.writer_dialect, k)}
        self.db.sync()
        self.db.close()
        if self.out_stream is not None:
            self.out_stream.close()


class ContextFile(object):
    def __init__(self, project_id, model_id, n_samples, keep_cols):
        hashable = reduce(operator.add, map(str,
                                            [project_id,
                                             model_id,
                                             n_samples,
                                             keep_cols]))
        digest = hashlib.md5(hashable.encode('utf8')).hexdigest()
        self.file_name = digest + '.shelve'

    def exists(self):
        """Does shelve exist. """
        return any(glob.glob(self.file_name + '*'))

    def clean(self):
        """Clean the shelve. """
        for fname in glob.glob(self.file_name + '*'):
            os.remove(fname)


class NewRunContext(RunContext):
    """RunContext for a new run.

    It creates a shelve file and adds a checkpoint journal.
    """

    def __enter__(self):
        if self.file_context.exists():
            self._ui.info('Removing old run shelve')
            self.file_context.clean()
        if os.path.exists(self.out_file):
            self._ui.warning('File {} exists.'.format(self.out_file))
            rm = self._ui.prompt_yesno('Do you want to remove {}'.format(
                self.out_file))
            if rm:
                os.remove(self.out_file)
                self._ui.info('Removed {}'.format(self.out_file))
            else:
                sys.exit(0)

        super(NewRunContext, self).__enter__()

        self.db['n_samples'] = self.n_samples
        self.db['project_id'] = self.project_id
        self.db['model_id'] = self.model_id
        self.db['keep_cols'] = self.keep_cols
        self.db['skip_row_id'] = self.skip_row_id
        self.db['output_delimiter'] = self.output_delimiter
        # list of batch ids that have been processed
        self.db['checkpoints'] = []
        # dictionary of error messages per batch
        self.db['warnings'] = {}
        self.db['errors'] = {}
        # used to check if output file is dirty (ie first write op)
        self.db['first_write'] = True
        self.db.sync()
        if six.PY2:
            self.out_stream = open(self.out_file, 'w+b')
        elif six.PY3:
            self.out_stream = open(self.out_file, 'w+', newline='')
        return self

    def __exit__(self, type, value, traceback):
        super(NewRunContext, self).__exit__(type, value, traceback)

    def batch_generator_args(self):
        """
        returns the arguments needed to set up a BatchGenerator
        In this case it's a fresh start
        """
        already_processed_batches = set()
        args = [self.dataset, self.n_samples, self.n_retry, self.delimiter,
                self._ui, self.fast_mode, self.encoding,
                already_processed_batches]
        return args


class OldRunContext(RunContext):
    """RunContext for a resume run.

    It requires a shelve file and plays back the checkpoint journal.
    Checks if inputs are consistent.

    TODO: add md5sum of dataset otherwise they might
    use a different file for resume.
    """

    def __enter__(self):
        if not self.file_context.exists():
            raise ValueError('Cannot resume a project without {}'
                             .format(self.FILENAME))
        super(OldRunContext, self).__enter__()

        if self.db['n_samples'] != self.n_samples:
            raise ShelveError('n_samples mismatch: should be {} but was {}'
                              .format(self.db['n_samples'], self.n_samples))
        if self.db['project_id'] != self.project_id:
            raise ShelveError('project id mismatch: should be {} but was {}'
                              .format(self.db['project_id'], self.project_id))
        if self.db['model_id'] != self.model_id:
            raise ShelveError('model id mismatch: should be {} but was {}'
                              .format(self.db['model_id'], self.model_id))
        if self.db['keep_cols'] != self.keep_cols:
            raise ShelveError('keep_cols mismatch: should be {} but was {}'
                              .format(self.db['keep_cols'], self.keep_cols))
        if self.db['skip_row_id'] != self.skip_row_id:
            raise ShelveError('skip_row_id mismatch: should be {} but was {}'
                              .format(self.db['skip_row_id'],
                                      self.skip_row_id))
        if self.db['output_delimiter'] != self.output_delimiter:
            raise ShelveError('output_delimiter mismatch: should be {}'
                              ' but was {}'.format(
                                    self.db['output_delimiter'],
                                    self.output_delimiter))

        self.db['warnings'] = {}
        self.db['errors'] = {}

        if six.PY2:
            self.out_stream = open(self.out_file, 'ab')
        elif six.PY3:
            self.out_stream = open(self.out_file, 'a', newline='')

        self._ui.info('resuming a shelved run with {} checkpointed batches'
                      .format(len(self.db['checkpoints'])))
        return self

    def __exit__(self, type, value, traceback):
        super(OldRunContext, self).__exit__(type, value, traceback)

    def batch_generator_args(self):
        """
        returns the arguments needed to set up a BatchGenerator
        In this case some batches may have already run
        """
        already_processed_batches = set(self.db['checkpoints'])
        args = [self.dataset, self.n_samples, self.n_retry, self.delimiter,
                self._ui, self.fast_mode, self.encoding,
                already_processed_batches]
        return args


def decode_writer_state(ch):
    return {
        b"-": "Initial",
        b"I": "Idle",
        b"G": "Getting from queue",
        b"D": "Done",
        b"W": "Writing",
    }.get(ch)


class WriterProcess(object):
    def __init__(self, ui, ctx, writer_queue, queue, deque, progress_queue,
                 abort_flag, writer_status, response_handlers):
        self._ui = ui
        self.ctx = ctx
        self.writer_queue = writer_queue
        self.queue = queue
        self.deque = deque
        self.progress_queue = progress_queue
        self.abort_flag = abort_flag
        self.response_handlers = response_handlers
        self.local_abort_flag = False
        self.writer_status = writer_status
        self.proc = None

    def deque_failed_batch(self, batch):
        # we retry a batch - decrement retry counter
        batch = batch._replace(rty_cnt=batch.rty_cnt - 1)
        try:
            self.deque.put(batch, block=True)
        except queue.Empty:
            self._ui.error('Dropping {} due to backfill queue full.'.format(
                batch))
            self.ctx.save_error(batch, error="Backfill queue full")

    def exit_fast(self, a, b):
        self.local_abort_flag = True

    def process_response(self):
        signal.signal(signal.SIGINT, self.exit_fast)
        signal.signal(signal.SIGTERM, self.exit_fast)

        """Process a successful request. """
        self._ui.debug('Writer Process started - {}'
                       ''.format(multiprocessing.current_process().name))

        rows_done = sum(rows for _, rows in self.ctx.db['checkpoints'])

        success = False
        processed = 0
        written = 0
        idle_cycles = 0
        last_report = time()

        try:
            while True:
                if self.abort_flag.value or self.local_abort_flag:
                    self._ui.debug('abort requested')
                    break
                try:
                    if idle_cycles > 2:
                        self.writer_status.value = b"I"
                    else:
                        self.writer_status.value = b"G"
                    msg, args = self.writer_queue.get(timeout=1)
                except queue.Empty:
                    idle_cycles += 1
                    continue
                idle_cycles = 0
                self.writer_status.value = b"W"

                if msg == WriterQueueMsg.CTX_ERROR:
                    # pred_name is a message if ERROR or WARNING
                    self._ui.debug('Writer ERROR')
                    self.ctx.save_error(args["batch"], error=args["error"])
                    continue
                elif msg == WriterQueueMsg.CTX_WARNING:
                    self._ui.debug('Writer WARNING')
                    self.ctx.save_warning(args["batch"], error=args["error"])
                    continue
                elif msg == WriterQueueMsg.SENTINEL:
                    self._ui.debug('Writer received SENTINEL')
                    break
                elif msg == WriterQueueMsg.RESPONSE:
                    processed += 1
                    batch = args["batch"]

                    unpack_data, format_data = self.response_handlers
                    try:
                        data, exec_time, elapsed_seconds = \
                            unpack_data(args['request'])
                        debug_msg = ('successful response {}-{}: exec time '
                                     '{:.0f}msec | round-trip: {:.0f}msec')
                        self._ui.debug(
                            debug_msg.format(batch.id, batch.rows, exec_time,
                                             elapsed_seconds * 1000))
                    except Exception as e:
                        self._ui.warning('{} response parse error: {} -- retry'
                                         ''.format(batch.id, e))
                        self.deque_failed_batch(batch)
                        continue

                    try:
                        written_fields, comb = format_data(
                            data, batch,
                            pred_name=self.ctx.pred_name,
                            pred_threshold_name=self.ctx.pred_threshold_name,
                            pred_decision_name=self.ctx.pred_decision_name,
                            keep_cols=self.ctx.keep_cols,
                            skip_row_id=self.ctx.skip_row_id,
                            fast_mode=self.ctx.fast_mode,
                            delimiter=self.ctx.dialect.delimiter,
                            max_prediction_explanations=(
                                self.ctx.max_prediction_explanations
                            )
                            )
                    except UnexpectedKeptColumnCount:
                        self._ui.fatal('Unexpected number of kept columns ' +
                                       'retrieved. This can happen in ' +
                                       '--fast mode with --keep_cols where ' +
                                       'some cells contain quoted delimiters')
                    except NoPredictionThresholdInResult:
                        self._ui.fatal('No predictionThreshold returned from '
                                       'API. --pred_threshold should be used '
                                       'only for binary classification '
                                       'predictions')
                    except Exception as e:
                        self._ui.fatal(e)

                    self.ctx.checkpoint_batch(batch, written_fields, comb)
                    written += 1
                    rows_done += batch.rows

                    if time() - last_report > REPORT_INTERVAL:
                        self.progress_queue.put((
                            ProgressQueueMsg.WRITER_PROGRESS, {
                                "processed": processed,
                                "written": written,
                                "rows": rows_done,
                                "rusage": get_rusage()
                            }))
                        last_report = time()
                else:
                    self._ui.error('Unknown Writer Queue msg: "{}", args={}'
                                   ''.format(msg, args))

            self._ui.debug('---Writer Exiting---')

            success = True
            if self.local_abort_flag:
                success = False

        except Exception as e:
            # Note this won't catch SystemExit which is raised by
            # sigterm_handler because it's based on BaseException
            self._ui.error('Writer Process error: batch={}, error={}'
                           ''.format(batch.id, e))

        finally:
            self.writer_status.value = b"D"
            self.progress_queue.put((ProgressQueueMsg.WRITER_DONE, {
                "ret": success,
                "processed": processed,
                "written": written,
                "rows": rows_done,
                "rusage": get_rusage()
            }))
            for o in [self.ctx, self.queue, self.writer_queue, self.deque,
                      self._ui]:
                if hasattr(o, 'close'):
                    # On Windows the Queue doesn't have a close attr
                    o.close()
            if success:
                sys.exit(0)
            else:
                sys.exit(1)

    def go(self):
        self._ui.set_next_UI_name('writer')
        self.ctx.close()      # handover it to Writer from top-level
        self.proc = \
            multiprocessing.Process(target=run_subproc_cls_inst,
                                    args=([self._ui, self.ctx,
                                           self.writer_queue,
                                           self.queue,
                                           self.deque,
                                           self.progress_queue,
                                           self.abort_flag,
                                           self.writer_status,
                                           self.response_handlers]),
                                    name='Writer_Proc')
        self.proc.start()
        return self.proc

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.proc and self.proc.is_alive():
            self._ui.debug('Terminating Writer')
            self.writer_queue.put_nowait((None, SENTINEL, None))


def run_subproc_cls_inst(_ui, ctx, writer_queue, queue, deque,
                         progress_queue, abort_flag, writer_status,
                         response_handlers):
    """
    This was intended to be a staticmethod of WriterProcess but
    python 2.7 on Windows can't find it unless it's at module level

    This should only be run in the spawned process to launch the
    WriterProcess class instance.
    """
    process_name = str(multiprocessing.current_process().name)
    if process_name != 'Writer_Proc':
        _ui.warning('WriterProcess.run_subproc_cls_inst called in '
                    'process named: "{}"'
                    ''.format(process_name))
    ctx.open()
    WriterProcess(_ui, ctx, writer_queue, queue, deque,
                  progress_queue, abort_flag, writer_status,
                  response_handlers).process_response()



import operator
from itertools import chain
import json
from six.moves import zip

from datarobot_batch_scoring.exceptions import UnexpectedKeptColumnCount, \
    NoPredictionThresholdInResult


def row_id_field(result_sorted, batch):
    """ Generate row_id field

    Parameters
    ----------
    result_sorted : list[dict]
        list of results sorted by rowId
    batch
        batch information

    Returns
    -------
    header: list[str]
    row_generator: iterator
    """
    return ['row_id'], (
        [record['rowId'] + batch.id]
        for record in result_sorted
    )


def prediction_values_fields(result_sorted, pred_name):
    """ Generate predictionValues fields

    Parameters
    ----------
    result_sorted : list[dict]
        list of results sorted by rowId
    pred_name: str, None
        column name for prediction results (may be None)

    Returns
    -------
    header: list[str]
    row_generator: iterator
    """
    single_row = result_sorted[0]
    fields = sorted(
        record['label']
        for record in single_row['predictionValues']
    )
    if pred_name is not None:
        headers = [pred_name]
        # batch scoring returns only positive class for binary tasks if
        # user specified pred_name option, so we eliminate negative
        # result
        if len(fields) == 2 and single_row['prediction'] in fields:
            fields = [fields[-1]]
    else:
        headers = fields

    rows_generator = (
        [
            pred_vals['value']
            for pred_vals in
            sorted(
                record['predictionValues'],
                key=operator.itemgetter('label')
            )
            if pred_vals['label'] in fields
        ] for record in result_sorted
    )

    return headers, rows_generator


def keep_original_fields(batch, keep_cols, fast_mode, input_delimiter):
    """ Generate fields from original data

    Parameters
    ----------
    batch
        batch of original data
    keep_cols: list[str]
        columns we should keep
    fast_mode: bool
        faster CSV processor. if True then row in the batch is raw string
    input_delimiter: str, optional
        csv delimiter. should be passed if fast_mode is True

    Returns
    -------
    header: list[str]
    row_generator: iterator
    """

    feature_indices = {col: i for i, col in
                       enumerate(batch.fieldnames)}
    indices = [feature_indices[col] for col in keep_cols]

    def rows_generator():
        for original_row in batch.data:
            if fast_mode:
                # row is a full line, we need to cut it into fields
                original_row = original_row.rstrip().split(input_delimiter)
                if len(original_row) != len(batch.fieldnames):
                    raise UnexpectedKeptColumnCount()

            yield [original_row[i] for i in indices]

    return keep_cols, rows_generator()


def prediction_explanation_fields(
    result_sorted,
    prediction_explanations_key,
    num_prediction_explanations,
):
    """ Generate prediction explanations fields

    Parameters
    ----------
    result_sorted : list[dict]
        list of results sorted by rowId
    prediction_explanations_key : str
        key of results by which explanation fields can be found
    num_prediction_explanations : int
        the total number of prediction explanations in response.
        Please note that in case output contains less explanations
        than this limit, the rest of the columns in the output will be pad
        with empty values for the sake of fixed columns count in the response.

    Returns
    -------
    header: list[str]
    row_generator: iterator
    """
    headers = []

    for num in range(1, num_prediction_explanations + 1):
        headers += [
            'explanation_{0}_feature'.format(num),
            'explanation_{0}_strength'.format(num)
        ]

    def rows_generator():
        for in_row in result_sorted:
            reason_codes = []
            for raw_reason_code in in_row[prediction_explanations_key]:
                reason_codes += [
                    raw_reason_code['feature'],
                    raw_reason_code['strength']
                ]
            # pad response with empty values in case not all prediction
            # explanations are present
            reason_codes += [''] * (len(headers) - len(reason_codes))
            yield reason_codes

    return headers, rows_generator()


def pred_threshold_field(result_sorted, pred_threshold_name):
    """ Generate prediction threshold field (for classification)
    Parameters
    ----------
    result_sorted : list[dict]
        list of results sorted by rowId
    pred_threshold_name : str
        column name which should contain prediction threshold
    Returns
    -------
    header: list[str]
    row_generator: iterator
    """

    return [pred_threshold_name], (
        [row.get('predictionThreshold')]
        for row in result_sorted
    )


def pred_decision_field(result_sorted, pred_decision):
    """ Generate prediction decision field

    Parameters
    ----------
    result_sorted : list[dict]
        list of results sorted by rowId
    pred_decision : str
        column name which should contain prediction decision (label)

    Returns
    -------
    header: list[str]
    row_generator: iterator
    """

    return [pred_decision], (
        [row.get('prediction')]
        for row in result_sorted
    )


def format_data(result, batch, **opts):
    """ Generate rows of response from results

    Parameters
    ----------
    result : list[dict]
        list of results
    batch
        batch information
    **opts
        additional arguments

    Returns
    -------
    written_fields : list
        headers
    comb : list
        list of rows
    """
    pred_name = opts.get('pred_name')
    pred_threshold_name = opts.get('pred_threshold_name')
    pred_decision_name = opts.get('pred_decision_name')
    keep_cols = opts.get('keep_cols')
    skip_row_id = opts.get('skip_row_id')
    fast_mode = opts.get('fast_mode')
    input_delimiter = opts.get('delimiter')
    max_prediction_explanations = opts.get('max_prediction_explanations')

    result_sorted = sorted(
        result,
        key=operator.itemgetter('rowId')
    )

    fields = []

    # region Columns

    # Row ID
    if not skip_row_id:
        fields.append(row_id_field(result_sorted, batch))

    # Kept columns from original data
    if keep_cols:
        fields.append(
            keep_original_fields(
                batch, keep_cols, fast_mode, input_delimiter
            )
        )

    # Prediction value columns
    fields.append(prediction_values_fields(result_sorted, pred_name))

    # Explanations columns
    single_row = result_sorted[0]
    prediction_explanations_keys = ('predictionExplanations', 'reasonCodes',
                                    None)

    def _find_prediction_explanations_key():
        return [key for key in prediction_explanations_keys
                if key in single_row or key is None][0]

    prediction_explanations_key = _find_prediction_explanations_key()
    if prediction_explanations_key:
        fields.append(
            prediction_explanation_fields(
                result_sorted,
                prediction_explanations_key,
                max_prediction_explanations,
            )
        )

    # Threshold field for classification
    # ('predictionThreshold' value from result)
    if pred_threshold_name:
        if 'predictionThreshold' in single_row:
            fields.append(
                pred_threshold_field(result_sorted, pred_threshold_name)
            )
        else:
            raise NoPredictionThresholdInResult()

    # Thresholded decision field ('prediction' value from result)
    if pred_decision_name:
        fields.append(pred_decision_field(result_sorted, pred_decision_name))

    # endregion

    headers = list(
        chain(*(header for header, _ in fields))
    )
    rows = list(
        list(chain(*row)) for row in
        zip(*(rows_gen for _, rows_gen in fields))
    )

    return headers, rows


def unpack_data(request):
    exec_time = float(request['headers']['X-DataRobot-Execution-Time'])
    result = json.loads(request['text'])  # replace with r.content
    elapsed_total_seconds = request['elapsed']
    return result['data'], exec_time, elapsed_total_seconds



import collections
import json
import logging
import multiprocessing
import signal
from functools import partial

from time import time
from concurrent.futures import FIRST_COMPLETED
from concurrent.futures import wait

from requests.utils import get_environ_proxies

try:
    from futures import ThreadPoolExecutor
except ImportError:
    from concurrent.futures import ThreadPoolExecutor

from six.moves import queue
import requests
import requests.adapters

from datarobot_batch_scoring.consts import (SENTINEL,
                                            REPORT_INTERVAL,
                                            ProgressQueueMsg, WriterQueueMsg)
from datarobot_batch_scoring.utils import get_rusage

from .base_network_worker import BaseNetworkWorker

TIMEOUT_WARNING = """\
The server did not send any data in the allotted amount of time.
 You might want to decrease the "--n_concurrent" parameters or increase
 "--timeout" parameter."""

logger = logging.getLogger(__name__)
FakeResponse = collections.namedtuple('FakeResponse', 'status_code, text')


class Network(BaseNetworkWorker):

    def _response_callback(self, r, batch=None, *args, **kw):
        try:
            if r.status_code == 200:
                pickleable_resp = {'elapsed': r.elapsed.total_seconds(),
                                   'text': r.text,
                                   'headers': r.headers}
                self.writer_queue.put((WriterQueueMsg.RESPONSE, {
                    "request": pickleable_resp,
                    "batch": batch
                }))
                return
            elif isinstance(r, FakeResponse):
                if r.status_code == 499:
                    msg = ('batch {} timed out, dropping; '
                           'we lost {} records'
                           ''.format(batch.id, len(batch.data)))
                    self.ui.error(msg)
                    self.send_error_to_ctx(batch, msg)
                    return

                self.ui.debug('Skipping processing response '
                              'because of FakeResponse')
            else:
                try:
                    fmt = 'batch {} failed with status code {} message: {}'
                    self.ui.warning(
                        fmt.format(batch.id, r.status_code,
                                   json.loads(r.text)['message'])
                    )
                except ValueError:
                    self.ui.warning('batch {} failed with status code: {}'
                                    ''.format(batch.id, r.status_code))

                text = r.text
                msg = ('batch {} failed, status_code:{} '
                       'text:{}'.format(batch.id, r.status_code, text))
                self.ui.error(msg)
                self.send_warning_to_ctx(batch, msg)

            if batch.rty_cnt == 1:
                msg = ('batch {} exceeded retry limit; '
                       'we lost {} records'
                       ''.format(batch.id, len(batch.data)))
                self.ui.error(msg)
                self.send_error_to_ctx(batch, msg)
            else:
                self.ui.warning('retrying failed batch {}, attempts left: {}'
                                .format(batch.id, batch.rty_cnt - 1))
                self.push_retry(batch)

        except Exception as e:
            msg = 'batch {} - dropping due to: {}, {} records lost'.format(
                batch.id, e, batch.rows)
            self.ui.error(msg)
            self.send_error_to_ctx(batch, msg)

    def push_retry(self, batch):
        # we retry a batch - decrement retry counter
        batch = batch._replace(rty_cnt=batch.rty_cnt - 1)
        try:
            self.network_deque.put(batch, block=False)
        except queue.Full:
            msg = 'Dropping {} due to backfill queue full.'.format(
                batch)
            self.ui.error(msg)
            self.send_error_to_ctx(batch, msg)

    def _request(self, request):

        try:
            prepared = self.session.prepare_request(request)
            proxies = get_environ_proxies(prepared.url)
            self.session.send(prepared, timeout=self._timeout, proxies=proxies)
        except Exception as exc:
            code = 400
            exc_tuple = (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            )
            is_timeout = isinstance(exc, exc_tuple)
            if is_timeout:
                self.ui.warning(TIMEOUT_WARNING)
                code = 499
            else:
                self.ui.debug('Exception {}: {}'.format(type(exc), exc))
                raise

            try:
                callback = request.kwargs['hooks']['response']
            except AttributeError:
                callback = request.hooks['response'][0]
            response = FakeResponse(code, 'No Response')
            callback(response)

    def request_cb(self, f):
        futures = [i for i in self.futures if not i.done()]
        self.ui.debug('request finished, pending futures: {}'
                      ''.format(len(futures)))

        if len(futures) == 0:
            self.state = b'E'

    def get_batch(self):
        while True:
            if self.abort_flag.value:
                self.exit_fast(None, None)
                break
            try:
                r = self.network_deque.get_nowait()
                self.ui.debug('Got batch from dequeu: {}'.format(r.id))
                self.n_retried += 1
                yield r
            except queue.Empty:
                try:
                    r = self.network_queue.get(timeout=1)
                    if r.id == SENTINEL.id:
                        break
                    self.n_consumed += 1
                    yield r
                except queue.Empty:
                    if self.state == b"E":
                        self.state = b'e'
                    elif self.state == b"e":
                        self.state = b'I'

                except OSError:
                    self.ui.error('OS Error')
                    break

    def perform_requests(self):
        signal.signal(signal.SIGINT, self.exit_fast)
        signal.signal(signal.SIGTERM, self.exit_fast)

        self.state = b'E'
        for q_batch in self.get_batch():
            for (batch, data) in self.split_batch(q_batch):

                hook = partial(self._response_callback, batch=batch)
                r = requests.Request(
                    method='POST',
                    url=self.endpoint,
                    headers=self.headers,
                    data=data,
                    auth=(self.user, self.api_token),
                    hooks={'response': hook})

                self.n_requests += 1

                while True:
                    self.futures = [i for i in self.futures if not i.done()]
                    if len(self.futures) < self.concurrency:
                        self.state = b'R'
                        f = self._executor.submit(self._request, r)
                        f.add_done_callback(self.request_cb)
                        self.futures.append(f)
                        break
                    else:
                        self.state = b'F'
                        wait(self.futures, return_when=FIRST_COMPLETED)
                yield
        #  wait for all batches to finish before returning
        self.state = b'W'
        while self.futures:
            f_len = len(self.futures)
            self.futures = [i for i in self.futures if not i.done()]
            if f_len != len(self.futures):
                self.ui.debug('Waiting for final requests to finish. '
                              'remaining requests: {}'
                              ''.format(len(self.futures)))
            wait(self.futures, return_when=FIRST_COMPLETED)
        self.state = b'D'
        yield True

    def run(self):
        self._executor = ThreadPoolExecutor(self.concurrency)
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            pool_connections=self.concurrency, pool_maxsize=self.concurrency)
        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)
        self.session.verify = self.verify_ssl

        t0 = time()
        last_report = time()
        i = 0
        r = None
        for r in self.perform_requests():
            if r is not True:
                i += 1
                self.ui.info('{} responses sent | time elapsed {}s'
                             .format(i, time() - t0))

                if time() - last_report > REPORT_INTERVAL:
                    self.progress_queue.put((
                        ProgressQueueMsg.NETWORK_PROGRESS, {
                            "processed": self.n_requests,
                            "retried": self.n_retried,
                            "consumed": self.n_consumed,
                            "rusage": get_rusage(),
                        }))
                    last_report = time()

        self.progress_queue.put((ProgressQueueMsg.NETWORK_DONE, {
            "ret": r,
            "processed": self.n_requests,
            "retried": self.n_retried,
            "consumed": self.n_consumed,
            "rusage": get_rusage(),
        }))

    def go(self):
        self.ui.set_next_UI_name('network')
        self.proc = \
            multiprocessing.Process(target=self.run,
                                    name='Netwrk_Proc')
        self.proc.start()
        return self.proc



import math
import re


class Detector(object):
    """
    Resample given data to use only full lines (w/o truncated).
    Detect possible delimiter character based on it's frequency. Gracefully
    handles multiline columns.
    """

    frequency_table = None
    non_delimiter_re = re.compile('[\w]', re.UNICODE)
    buff_size = 1024 * 2048

    def __init__(self):
        self.frequency_table = {}

    def increment(self, char, lines):
        if char not in self.frequency_table:
            self.frequency_table[char] = {}
        if lines not in self.frequency_table[char]:
            self.frequency_table[char][lines] = 0

        self.frequency_table[char][lines] += 1

    def detect(self, sample, quotechar='"'):
        lines_analyzed, lines_content = self.get_sample(sample,
                                                        quotechar=quotechar)
        resampled = "".join(lines_content[:-1])
        return self.analyze(lines_analyzed), resampled

    def get_sample(self, sample, sample_lines=20, quotechar='"'):
        """
        Given actual sample method extracts complete lines and fills delimiter
        candidates table with frequencies of occurrences.
        """
        enclosed = False
        actual_lines = 1
        lines = []
        single_line = ""

        for idx, ch in enumerate(sample):
            if actual_lines >= sample_lines:
                break
            prev_ch = sample[idx-1] if idx else None
            next_ch = sample[idx+1] if (idx + 1) < len(sample) else None
            single_line += ch

            if ch == quotechar:
                if enclosed and next_ch != quotechar:
                    enclosed = False
                elif not enclosed:
                    enclosed = True
            elif not enclosed and \
                    (ch == '\n' and prev_ch != '\r' or ch == '\r'):
                actual_lines += 1
                lines.append(single_line)
                single_line = ""

            elif not enclosed and not self.non_delimiter_re.match(ch):
                self.increment(ch, actual_lines)

        return actual_lines, lines

    def analyze(self, lines_analyzed):
        candidates = []
        for delim, freq in self.frequency_table.items():
            deviation = self.deviation(freq, lines_analyzed)

            if float(0.0) == deviation:
                candidates.append(delim)
        return candidates

    def mean(self, line_freq, lines_analyzed):
        """ Calculates mean frequency of potential delimiters occurrences

        :param line_freq: dict[int]int
        :param lines_analyzed: int
        :return: int
            mean value of all characters appeared at sample
        """
        freqs = list(line_freq.values())[:lines_analyzed]
        return sum(freqs) / lines_analyzed

    def deviation(self, line_freq, lines_analyzed):
        lines_to_account = lines_analyzed - 1

        average = self.mean(line_freq, lines_to_account)
        frequencies = list(line_freq.values())[:lines_to_account]

        squares = map(lambda frequency: (float(average) - frequency) ** 2,
                      frequencies)

        try:
            deviation = math.sqrt(sum(squares) / (lines_to_account - 1))
        except ZeroDivisionError:
            deviation = -1

        return deviation



#!/usr/bin/env python
import sys
import codecs
import os.path
import re
from setuptools import setup, find_packages

extra = {}

def read_requirements_file(file):
    fname = os.path.join(os.path.abspath(os.path.dirname(__file__)), file)
    with open(fname, 'r') as r:
        return r.readlines()


install_requires = read_requirements_file('requirements-base.txt')

# Since futures 3.2 [1], the package enforces to be installed only in Python 2
# environments because it's basically a backport of Python 3's built-in
# package. So in order to support both Python 2 and Python 3 environments, we
# have to skip installation of futures package in case of Python 3.
#
# It might look natural to use environment markers [2] to achieve this goal but
# they are new and were introduced in setuptools in mid of 2017. FWIW,
# setuptools on both Ubuntu Trusty and Ubuntu Xenial do not support them and
# batch scoring script may be installed in pretty outdated envs. So let's do it
# old-fashioned way by adding condition here.
#
# The above is implemented by splitting dependencies into 2 files:
# `requirements-base.txt` - common deps for Py3 and Py2
# `requirements-py27.txt` - for Python 2 only
#
# [1] https://github.com/agronholm/pythonfutures/commit/d0393ad626d25622927bb0ed47d35ddb2f6cd321 # noqa: E501
# [2] https://www.python.org/dev/peps/pep-0508/#environment-markers

if sys.version_info[0] < 3:
    install_requires.extend(
        read_requirements_file('requirements-py27.txt')
    )

extra['entry_points'] = {
    'console_scripts': [
        'batch_scoring = datarobot_batch_scoring.main:main',
        'batch_scoring_sse = datarobot_batch_scoring.main:main_standalone',
        'batch_scoring_deployment_aware = datarobot_batch_scoring.main:main_deployment_aware'
    ]}
extra['install_requires'] = install_requires


this_directory = os.path.abspath(os.path.dirname(__file__))


init_fname = os.path.join(this_directory, 'datarobot_batch_scoring', '__init__.py')
with codecs.open(init_fname, 'r', 'latin1') as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


readme_fname = os.path.join(this_directory, 'README.rst')
with codecs.open(readme_fname, 'r', 'utf-8') as f:
    long_description = f.read()


setup(
    name='datarobot_batch_scoring',
    version=version,
    description=("A script to score CSV files via DataRobot's prediction API"),
    long_description=long_description,
    author='DataRobot',
    author_email='support@datarobot.com',
    maintainer='DataRobot',
    maintainer_email='support@datarobot.com',
    license='BSD',
    url='http://www.datarobot.com/',
    packages=find_packages(),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    **extra
)
